from __future__ import print_function

import logging
import importlib as IM

import six
import rdflib as R

from .context_mapped_class_util import find_base_namespace

L = logging.getLogger(__name__)


class MappedClass(type):

    """A type for MappedClasses

    Sets up the graph with things needed for MappedClasses
    """
    def __init__(self, name, bases, dct):
        L.debug("INITIALIZING %s", name)
        super(MappedClass, self).__init__(name, bases, dct)

        self.__base_namespace = find_base_namespace(dct, bases)

        self.__rdf_type = None
        # Set the rdf_type early
        if 'rdf_type' in dct:
            self.__rdf_type = R.URIRef(dct['rdf_type'])

        if self.__rdf_type is None:
            self.__rdf_type = self.base_namespace[self.__name__]

        self.__rdf_namespace = None
        rdf_ns = dct.get('rdf_namespace', None)
        if rdf_ns is not None:
            L.debug("Setting rdf_namespace to {}".format(rdf_ns))
            if not isinstance(rdf_ns, R.Namespace):
                rdf_ns = R.Namespace(rdf_ns)
            self.__rdf_namespace = rdf_ns

        self.__schema_namespace = None
        schema_ns = dct.get('schema_namespace', None)
        if schema_ns is not None:
            L.debug("Setting schema_namespace to {}".format(schema_ns))
            if not isinstance(schema_ns, R.Namespace):
                schema_ns = R.Namespace(schema_ns)
            self.__schema_namespace = schema_ns

        if self.__rdf_namespace is None:
            if self.base_data_namespace is not None:
                rdf_namespace = self.base_data_namespace[self.__name__] + '#'
            else:
                rdf_namespace = self.base_namespace[self.__name__] + '#'
            L.debug("Setting rdf_namespace to {}".format(rdf_namespace))
            self.__rdf_namespace = R.Namespace(rdf_namespace)

        if self.__schema_namespace is None:
            L.debug("Setting schema_namespace to {}".format(self.base_namespace[self.__name__] + '/'))
            self.__schema_namespace = R.Namespace(
                self.base_namespace[self.__name__] + "/")

        self.__rdf_type_object_callback = dct.get('rdf_type_object_callback')
        self.__rdf_type_object = dct.get('rdf_type_object')

        if not getattr(self, 'unmapped', False) and not dct.get('unmapped'):
            self.register_on_module()

    def register_on_module(self, module=None):
        '''
        "Registers" this class on a module (typically the one in which the class is
        defined) such that owmeta-core functions can locate it. This happens automatically
        when the class is defined unless the 'unmapped' attribute is defined and set to
        `True`.

        This mechanism necessary in some cases where classes are generated dynamically or
        in a method and aren't necessarily assigned to attributes on the module where they
        are defined.
        '''
        module = module or IM.import_module(self.__module__)
        if not hasattr(module, '__yarom_mapped_classes__'):
            module.__yarom_mapped_classes__ = [self]
        else:
            module.__yarom_mapped_classes__.append(self)

    @property
    def base_namespace(self):
        return self.__base_namespace

    @property
    def rdf_type_object(self):
        if self.__rdf_type_object_callback is not None:
            rdto = self.__rdf_type_object_callback()
            if rdto is not None:
                self.__rdf_type_object_callback = None
                self.__rdf_type_object = rdto
        return self.__rdf_type_object

    @rdf_type_object.setter
    def rdf_type_object(self, value):
        if value is not None:
            self.__rdf_type_object_callback = None
        self.__rdf_type_object = value

    @property
    def rdf_type(self):
        return self.__rdf_type

    @rdf_type.setter
    def rdf_type(self, new_type):
        if not isinstance(new_type, R.URIRef) and \
                isinstance(new_type, (str, six.text_type)):
            new_type = R.URIRef(new_type)
        self.__rdf_type = new_type

    @property
    def rdf_namespace(self):
        return self.__rdf_namespace

    @property
    def schema_namespace(self):
        return self.__schema_namespace

    def __lt__(self, other):
        res = False
        if issubclass(other, self) and not issubclass(self, other):
            res = True
        elif issubclass(self, other) == issubclass(other, self):
            res = self.__name__ < other.__name__
        return res

    def on_mapper_add_class(self, mapper):
        """ Called by :class:`owmeta_core.mapper.Mapper`

        Registers certain properties of the class
        """
        return self



from git import Repo


class GitRepoProvider(object):

    def __init__(self):
        self._repo = None
        self.base = None

    def init(self, base=None):
        base = self.base if not base else base
        self._repo = Repo.init(base)

    def add(self, files):
        self.repo().index.add(files)

    def remove(self, files, recursive=False):
        self.repo().index.remove(files, r=recursive)

    def reset(self):
        from git.refs.head import HEAD
        HEAD(self.repo()).reset(working_tree=True)

    def commit(self, msg):
        self.repo().index.commit(msg)

    def repo(self):
        if self._repo is None:
            self._repo = Repo(self.base)
        return self._repo

    def clone(self, url, base, progress=None, **kwargs):
        if progress is not None:
            progress = _CloneProgress(progress)
        Repo.clone_from(url, base, progress=progress, **kwargs)

    @property
    def is_dirty(self):
        return self.repo().is_dirty()


class _CloneProgress(object):

    def __init__(self, progress_reporter):
        self.pr = progress_reporter
        try:
            self.pr.unit = 'objects'
        except AttributeError:
            pass

        self._opcode = 0

    def __call__(self, op_code, cur_count, max_count=None, message=''):
        if op_code != self._opcode:
            self.pr.n = 0
        if max_count is not None:
            self.pr.total = max_count
        self.pr.update(cur_count - self.pr.n)



# -*- coding: utf-8 -*-

"""
.. _owm_module:

owmeta_core
===========
owmeta-core is a platform for sharing relational data over the internet.
"""

from __future__ import print_function
__version__ = '0.14.0.dev20210525235610'
__author__ = 'OpenWorm.org authors and contributors'

import sys
import os
import logging
import uuid
from os.path import join as pth_join

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

BASE_SCHEMA_URL = 'http://schema.openworm.org/2020/07'
BASE_DATA_URL = 'http://data.openworm.org'

# The c extensions are incompatible with our code...
os.environ['WRAPT_DISABLE_EXTENSIONS'] = '1'


OWMETA_PROFILE_DIR = os.environ.get('OWMETA_PROFILE_DIR', pth_join('~', '.owmeta'))
'''
Base directory in the user's profile for owmeta (e.g., shared configuration, bundle cache)
'''


from .configure import Configurable
from .context import Context, ClassContext

__all__ = [
    "get_data",
    "disconnect",
    "connect",
    "Configurable",
]

DEF_CTX = Context()

RDF_CONTEXT = ClassContext(ident='http://www.w3.org/1999/02/22-rdf-syntax-ns',
        base_namespace='http://www.w3.org/1999/02/22-rdf-syntax-ns#')

RDFS_CONTEXT = ClassContext(ident='http://www.w3.org/2000/01/rdf-schema',
        imported=(RDF_CONTEXT,),
        base_namespace='http://www.w3.org/2000/01/rdf-schema#')

BASE_CONTEXT = ClassContext(imported=(RDFS_CONTEXT,),
        ident=BASE_SCHEMA_URL,
        base_namespace=BASE_SCHEMA_URL + '#')


def get_data(path):
    # get a resource from the installed package location

    from sysconfig import get_path
    from pkgutil import get_loader
    from glob import glob
    package_paths = glob(os.path.join(get_path('platlib'), '*'))
    sys.path = package_paths + sys.path
    installed_package_root = os.path.dirname(get_loader('owmeta_core').get_filename())
    sys.path = sys.path[len(package_paths):]
    filename = os.path.join(installed_package_root, path)
    return filename


class Connection(object):
    '''
    Connection to an owmeta_core database. Essentially, wraps a `~owmeta_core.data.Data`
    object.
    '''

    def __init__(self, configFile=None, conf=None, mapper=None):
        """
        Load desired configuration and open the database

        Parameters
        ----------
        configFile : str, optional
            The configuration file for owmeta_core.
        conf : dict, .Configuration, .Data, optional
            A configuration object for the connection. Takes precedence over `configFile`
        mapper : owmeta_core.mapper.Mapper
            Provides the mapper for this connection

        Returns
        -------
        Connection
            connection wrapping the configuration
        """
        from .data import Data, DatabaseConflict
        from .mapper import Mapper

        if configFile is not None and not isinstance(configFile, str):
            conf = configFile
            configFile = None

        if conf:
            if not isinstance(conf, Data):
                conf = Data(conf)
        elif configFile:
            conf = Data.open(configFile)
        else:
            conf = Data({"rdf.source": "default"})

        try:
            conf.init_database()
        except DatabaseConflict as e:
            raise ConnectionFailError(e, "It looks like a connection is already opened by a living process")
        except Exception as e:
            raise ConnectionFailError(e)

        logging.getLogger('owmeta_core').info("Connected to database")

        self.conf = conf

        if mapper is None:
            mapper = Mapper(conf=conf)

        self._context = Context(conf=self.conf, mapper=mapper)

        self.identifier = str(uuid.uuid4())
        '''
        Identifier for this connection.

        Primarily, so that this Connection can be passed to contextualize for a Context
        '''

        self.mapper = mapper

    @property
    def rdf(self):
        return self.conf['rdf.graph']

    def disconnect(self):
        '''
        Close the database and stop listening to module loaders
        '''
        self.conf.closeDatabase()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.disconnect()

    def __call__(self, target):
        '''
        Contextualize the given `Context`
        '''
        if target is not None:
            return target.contextualize(self._context)
        else:
            raise TypeError('Connections can only contextualize owmeta_core.context.Context'
                    ' or subclasses thereof. Received %s' % target)

    def __str__(self):
        conf = self.conf
        return 'Connection:{source}:{store_conf}'.format(
                source=conf.get('rdf.source'),
                store_conf=conf.get('rdf.store_conf', 'default'))


def disconnect(c=None):
    """ Close the connection. """
    if c:
        c.disconnect()


class ConnectionFailError(Exception):
    '''
    Thrown when a connection fails
    '''
    def __init__(self, cause, *args):
        if args:
            super(ConnectionFailError, self).__init__('owmeta_core connection failed: {}. {}'.format(cause, *args))
        else:
            super(ConnectionFailError, self).__init__('owmeta_core connection failed: {}'.format(cause))


connect = Connection



from .capability import Capability, Provider


class FilePathCapability(Capability):
    '''
    Provides a file path where named files can be retrieved.

    This capability may be needed when files are referred to that aren't necessarily
    stored on the local machine, or which on the local machine, but only in non-portable
    locations (e.g., a home directory).
    '''


class FilePathProvider(Provider):
    '''
    Provides the `FilePathCapability`
    '''

    provided_capabilities = [FilePathCapability()]

    def file_path(self):
        '''
        The needed file path
        '''
        raise NotImplementedError()


class CacheDirectoryCapability(Capability):
    '''
    Capability that provides a cache directory.

    The provider of this capability must be capable of persisting effectively distinct
    directories for each `Capable` which needs this capability. The provider must permit
    depositing files in the directory by the current effective user.
    '''


class CacheDirectoryProvider(Provider):
    '''
    Provides the `CacheDirectoryCapability`
    '''

    provided_capabilities = [CacheDirectoryCapability()]

    def clear(self, cache_key):
        '''
        Clear the cache directory for the `Capable`.

        Should remove the directory itself, if possible.
        '''
        raise NotImplementedError

    def cache_directory(self, cache_key):
        '''
        Return the cache directory path

        Parameters
        ----------
        cache_key : str
            The key for the cache entry

        Returns
        -------
        str
            The cache directory
        '''
        raise NotImplementedError


# Possible other capabilities:
# - http/socks proxy
# - user name / password



'''
A replpacement for numpydoc's docscrape that doesn't require numpydoc's time-consuming
imports
'''
import re
from textwrap import dedent
from collections import namedtuple

parameter_regex = r'''
^(?P<param_name>[*]{0,2}\w+)(?:\s+:\s+(?P<param_type>.+))?\n
(?P<param_description>(^[ ]{4}(?:(\S.*|)\n))+)
'''

regex = r'''
(?P<desc>(?:(^\S.*)?\n)+(?:^\n+(?=Parameters)))?
(^Parameters\n
^-+\n
(?P<parameters>(?:{parameter_regex})+))?
'''.format(parameter_regex=parameter_regex)

first_line_unindented_regex = r'^\S.*\n+(?P<initial_white_space>\s+)'

RE = re.compile(regex, flags=re.VERBOSE | re.MULTILINE)
ParamRE = re.compile(parameter_regex, flags=re.VERBOSE | re.MULTILINE)
FLURE = re.compile(first_line_unindented_regex, flags=re.VERBOSE | re.MULTILINE)


ParamInfo = namedtuple('ParamInfo', ('name', 'val_type', 'desc'))


def parse(text):
    resp = {}
    iws_match = FLURE.match(text)
    if iws_match:
        text = iws_match.group('initial_white_space') + text
    if text.startswith('\n'):
        text = text[1:]
    text = dedent(text)
    md = RE.match(text)
    if md:
        desc = md.group('desc')
        resp['desc'] = desc and desc.strip()
        resp['parameters'] = []
        params = md.group('parameters')
        if params:
            for pmd in ParamRE.finditer(params):
                param_type = pmd.group('param_type')
                tp = ParamInfo(pmd.group('param_name').strip(),
                               param_type and param_type.strip(),
                               pmd.group('param_description').strip())
                resp['parameters'].append(tp)
    if not resp.get('desc') and not resp.get('parameters'):
        resp['desc'] = text.strip()
    return resp



from json import loads, dumps

from rdflib.term import Literal, bind, Identifier, URIRef
import six

from .quantity import Quantity
from .utils import FCN
from . import BASE_SCHEMA_URL
bind(URIRef(BASE_SCHEMA_URL + '/datatype/quantity'),
     Quantity,
     constructor=Quantity.parse,
     lexicalizer=Quantity.__str__)

bind(URIRef(BASE_SCHEMA_URL + '/datatype/list'),
     list,
     constructor=loads,
     lexicalizer=dumps)
bind(URIRef(BASE_SCHEMA_URL + '/datatype/object'),
     dict,
     constructor=loads,
     lexicalizer=lambda x: dumps(x, sort_keys=True))

# XXX: RDFLib should take the first match for a python type, so we'll translate into the
# list and object types above
bind(URIRef('http://markw.cc/yarom/schema/datatype/list'),
     list,
     constructor=loads,
     lexicalizer=dumps)
bind(URIRef('http://markw.cc/yarom/schema/datatype/object'),
     dict,
     constructor=loads,
     lexicalizer=lambda x: dumps(x, sort_keys=True))


class PropertyValue(object):
    """ Holds a literal value for a property """

    # Made to look like a GraphObject, but not subclassing so we can have slots

    __slots__ = ('value', 'properties', 'owner_properties')

    def __init__(self, value):
        self.properties = []
        self.owner_properties = []
        if not isinstance(value, Identifier):
            self.value = Literal(value)
        else:
            self.value = value

    def triples(self, *args, **kwargs):
        return []

    @property
    def identifier(self):
        return self.value

    @property
    def defined(self):
        return True

    @property
    def idl(self):
        return self.identifier

    def __hash__(self):
        return hash(self.value)

    def __str__(self):
        if six.PY3:
            return "PV("+self.value+")"
        else:
            return "PV("+self.value.encode('UTF-8')+")"

    def __repr__(self):
        return '{}({})'.format(FCN(type(self)), repr(self.value))

    def __lt__(self, other):
        return self.value < other.value

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif isinstance(other, PropertyValue):
            return self.value == other.value
        else:
            return False
